@@ -1,26 +1,74 @@
-#include "pybind11/pybind11.h"
+//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+//
+// lsdtopytools - main 
+// Land Surface Dynamics python binding
+//
+// An object within the University
+//  of Edinburgh Land Surface Dynamics group topographic toolbox
+//  for Using the LSDTopoTools framwork from python without having to compile manually the code
+//  and to script it. Cross platform, this requires a bit of cpp-python gymnastic using pybind11 and xtensor.
+//  The original version of that file was created from a template from python-binding cookiecutter, from xtensor git-hub. 
+//
+// Developed by:
+//  Boris Gailleton
+//  Simon M. Mudd
+//
+// Copyright (C) 2018 Simon M. Mudd 2018
+//
+// Developer can be contacted by simon.m.mudd _at_ ed.ac.uk
+//
+//    Simon Mudd
+//    University of Edinburgh
+//    School of GeoSciences
+//    Drummond Street
+//    Edinburgh, EH8 9XP
+//    Scotland
+//    United Kingdom
+//
+// This program is free software;
+// you can redistribute it and/or modify it under the terms of the
+// GNU General Public License as published by the Free Software Foundation;
+// either version 2 of the License, or (at your option) any later version.
+//
+// This program is distributed in the hope that it will be useful,
+// but WITHOUT ANY WARRANTY;
+// without even the implied warranty of
+// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+// See the GNU General Public License for more details.
+//
+// You should have received a copy of the
+// GNU General Public License along with this program;
+// if not, write to:
+// Free Software Foundation, Inc.,
+// 51 Franklin Street, Fifth Floor,
+// Boston, MA 02110-1301
+// USA
+//
+//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 
+// These includes are for binding with python
+#include "pybind11/pybind11.h"
 #include "xtensor/xmath.hpp"
 #include "xtensor/xarray.hpp"
-
 #define FORCE_IMPORT_ARRAY
 #include "xtensor-python/pyarray.hpp"
 #include "xtensor-python/pytensor.hpp"
 #include "xtensor-python/pyvectorize.hpp"
 #include "xtensor/xadapt.hpp"
-
-#include <iostream>
-#include <numeric>
-#include <cmath>
-#include <vector>
-#include <map>
-
 #include <pybind11/stl_bind.h>
 #include <pybind11/stl.h>
 #include <pybind11/complex.h>
 #include <pybind11/chrono.h>
 #include <pybind11/functional.h>
 
+// Common includes
+#include <iostream>
+#include <numeric>
+#include <cmath>
+#include <vector>
+#include <map>
+
+// Includes for openMP (experimental use, you can ignore)
 #include <omp.h>  //Used for OpenMP run-time functions
 #ifndef _OPENMP
   #define omp_get_thread_num()  0
@@ -28,17 +76,24 @@
   #define omp_get_max_threads() 1
 #endif
 
+// Other lsdtopytools includes
 #include "LSD_xtensor_utils.hpp"
 #include "LSDStatsTools.hpp"
 #include "LSDEntry_points.hpp"
 #include "LSD_xtensor_convtools.hpp"
-// #include "LSDTribBasModel.hpp"
+// #include "LSDTribBasModel.hpp" // Moved to Carpythians
 #include "LSDDEM_xtensor.hpp"
-
-
-
 namespace py = pybind11;
 
+
+
+// ###########################################################################################################
+// ###########################################################################################################
+// ############################### First attempts to bind, ignore these inline functions #####################
+// ################################## I keep them for back-compatibility purposes  ###########################
+// ###########################################################################################################
+// ###########################################################################################################
+
 inline std::map<int, std::map<std::string,float> > comparison_stats_from_2darrays(xt::pytensor<int, 2>& arr1, xt::pytensor<float, 2>& arr2, float ignore_value, int nRows, int nCols)
 {py::gil_scoped_release release; std::map<int, std::map<std::string,float> > output = xtlsd::_comparison_stats_from_2darrays(arr1,arr2,ignore_value,nRows,nCols); return output;}
 
@@ -69,8 +124,6 @@ inline std::map<std::string, xt::pytensor<float,2> > get_polyfit_rasters(int nro
 inline std::map<std::string, xt::pytensor<float,1> > get_circular_windowed_stats(int nrows, int ncols, float xmin, float ymin, float cellsize, float ndv, xt::pytensor<float,2>& data, xt::pytensor<float,1>& x_coord, xt::pytensor<float,1>& y_coord , float radius)
 { return EPPy::get_circular_windowed_stats(nrows, ncols, xmin, ymin, cellsize, ndv, data, x_coord, y_coord , radius);}
 
-//#####################################################################################################
-
 // The follwing function is just there for debugging purpose
 inline std::vector<float> compt_debug(xt::pytensor<float,1>& x_val)
 {
@@ -114,27 +167,38 @@ inline std::vector<float> compt_debug(xt::pytensor<float,1>& x_val)
   return v;
 
 }
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//######################################################################## Find below the actual bindings ##################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
+//##########################################################################################################################################################################################################
 
 
 // Python Module and Docstrings
-
 PYBIND11_MODULE(lsdtt_xtensor_python, m)
 {
     xt::import_numpy();
 
     m.doc() = R"pbdoc(
-        This is an attempt to port LSDTT to Python
+        lsdtt_xtensor_python:
+        Python bindings for LSDTopoTools
 
         .. currentmodule:: lsdtt_xtensor_python
 
         .. autosummary::
-           :toctree: _generate
-
-          comparison_stats_from_2darrays
-          get_groupped_values
-          KDE_gaussian
-          compt_debug
-          growing_window_stat       
+           :toctree: _generate  
            
     )pbdoc";
 
@@ -151,6 +215,7 @@ PYBIND11_MODULE(lsdtt_xtensor_python, m)
     m.def("get_polyfit_rasters",get_polyfit_rasters, "Gets polyfit rasters with parameters that I will develop later. SNS.");
     m.def("get_circular_windowed_stats",get_circular_windowed_stats, "Gets polyfit rasters with parameters that I will develop later. SNS.",py::call_guard<py::gil_scoped_release>());
 
+    m.def("segment_data_Mudd14", &mudd14partitioner::segment_data);
 
     py::class_<LSDDEM_xtensor>(m, "LSDDEM_cpp",py::dynamic_attr())
       .def(py::init<>())
@@ -179,7 +244,7 @@ PYBIND11_MODULE(lsdtt_xtensor_python, m)
       .def("get_PP_raster", &LSDDEM_xtensor::get_PP_raster)
       .def("get_catchment_perimeter", &LSDDEM_xtensor::get_catchment_perimeter)
       .def("get_base_raster", &LSDDEM_xtensor::get_base_raster)
-      .def("calculate_movern_disorder", &LSDDEM_xtensor::calculate_movern_disorder)
+      .def("calculate_movern_disorder", &LSDDEM_xtensor::calculate_movern_disorder, py::call_guard<py::gil_scoped_release>())
       .def("get_disorder_dict", &LSDDEM_xtensor::get_disorder_dict)
       .def("get_disorder_vec_of_tested_movern", &LSDDEM_xtensor::get_disorder_vec_of_tested_movern)
       .def("burn_rast_val_to_xy", &LSDDEM_xtensor::burn_rast_val_to_xy)
@@ -199,6 +264,26 @@ PYBIND11_MODULE(lsdtt_xtensor_python, m)
       .def("mask_topo", &LSDDEM_xtensor::mask_topo)
       .def("get_individual_basin_raster", &LSDDEM_xtensor::get_individual_basin_raster)
       .def("get_best_fits_movern_per_BK", &LSDDEM_xtensor::get_best_fits_movern_per_BK)
+      .def("set_n_nodes_to_visit_downstream", &LSDDEM_xtensor::set_n_nodes_to_visit_downstream)
+      .def("get_single_river_from_top_to_outlet", &LSDDEM_xtensor::get_single_river_from_top_to_outlet)
+      .def("query_rowcol_from_xy", &LSDDEM_xtensor::query_rowcol_from_xy)
+      .def("get_receiver_data", &LSDDEM_xtensor::get_receiver_data)
+      .def("get_disorder_dict_SK", &LSDDEM_xtensor::get_disorder_dict_SK)
+      .def("get_SA_from_vertical_interval",&LSDDEM_xtensor::get_SA_from_vertical_interval)
+      .def("trim_nodes_draining_to_baselevel", &LSDDEM_xtensor::trim_nodes_draining_to_baselevel)
+      .def("calculate_outlets_locations_from_xy_v2", &LSDDEM_xtensor::calculate_outlets_locations_from_xy_v2)
+      .def("extract_basin_draining_to_coordinates", &LSDDEM_xtensor::extract_basin_draining_to_coordinates)
+      .def("calculate_outlets_min_max_draining_to_baselevel", &LSDDEM_xtensor::calculate_outlets_min_max_draining_to_baselevel)
+      .def("query_xy_for_each_basin", &LSDDEM_xtensor::query_xy_for_each_basin)
+      .def("get_DEBUG_all_dis_SK",&LSDDEM_xtensor::get_DEBUG_all_dis_SK)
+      .def("get_all_disorder_values",&LSDDEM_xtensor::get_all_disorder_values)
+      .def("get_n_pixels_by_combinations", &LSDDEM_xtensor::get_n_pixels_by_combinations)
+      .def("calculate_outlet_location_of_main_basin", &LSDDEM_xtensor::calculate_outlet_location_of_main_basin)
+      .def("calculate_outlets_locations_from_range_of_DA", &LSDDEM_xtensor::calculate_outlets_locations_from_range_of_DA)
+      .def("get_baselevels_XY", &LSDDEM_xtensor::get_baselevels_XY)
+      .def("get_chi_raster_all", &LSDDEM_xtensor::get_chi_raster_all)
+      .def("get_flow_distance_raster",&LSDDEM_xtensor::get_flow_distance_raster)
+  
       ;// end of class LSDDEM (semicolon important)
 
     // TribBas
